port = 8000 
upload_dir = 'web'
hot_reload = True
appname = 'LLM test'
autotest = '*'
#lm = 'host', "qwen3.5:9b" ,"http://localhost:11434/v1"
#llm = 'groq', 'llama-3.1-70b-versatile'
llm = 'google', "gemini-2.5-flash"
#llm = 'host', "https://openrouter.ai/api/v1", "OPENROUTER1_API_KEY", "deepseek/deepseek-chat-v3-0324:free"
#llm = 'openai', "gpt-4o"
#llm = 'mistral', "mistral-large-latest"
#llm_cache = 'llm.cache'
