#header[data-v-69aea6da]{font-size:16px;font-weight:700;left:10px;pointer-events:none;position:absolute;top:10px;z-index:2}#graph[data-v-69aea6da]{border:1px solid #ccc;height:600px;position:relative;width:100%}