"use strict";(globalThis["webpackChunkuniqua"]=globalThis["webpackChunkuniqua"]||[]).push([[609],{1609:(e,t,n)=>{n.r(t),n.d(t,{default:()=>b});var l=n(1758);const o={class:"fullscreen bg-blue text-white text-center q-pa-md flex flex-center"};function s(e,t,n,s,c,a){const r=(0,l.g2)("q-btn");return(0,l.uX)(),(0,l.CE)("div",o,[(0,l.Lk)("div",null,[t[0]||(t[0]=(0,l.Lk)("div",{style:{"font-size":"30vh"}}," 404 ",-1)),t[1]||(t[1]=(0,l.Lk)("div",{class:"text-h2",style:{opacity:".4"}}," Oops. Nothing here... ",-1)),(0,l.bF)(r,{class:"q-mt-xl",color:"white","text-color":"blue",unelevated:"",to:"/",label:"Go Home","no-caps":""})])])}const c=(0,l.pM)({name:"Error404"});var a=n(2807),r=n(6384),u=n(8582),i=n.n(u);const h=(0,a.A)(c,[["render",s]]),b=h;i()(c,"components",{QBtn:r.A})}}]);